// SRP Server Helper for JAVA
// help_srp_server.c
// Created: 12/3/99
// Modified: 12/3/99

// 12-3  mpiecuch     Initial Creation 

// Compile: gcc srpserver.c


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include "srpserverlib.c"


int main(int argc, char **argv) {
  // Testing main function
  int udpsockno;
  int srpsock;
  int i;
  char test_data[255];


  udpsockno = helper_setup_udp_server_socket(55555);
  srpsock = helper_start_srp_server(udpsockno);
  
  for (i=0;i<20;i++) {
    sprintf(test_data, "Test Data: %d", i);
    printf("Sending %d\n", i);
    helper_sendto_srp(srpsock, (int) test_data, strlen(test_data) + 1);
    sleep(10);
  }
  helper_stop_server_srp(srpsock);

}

int helper_setup_udp_server_socket (int port) {

  struct sockaddr_in localaddr;
  int size_localaddr;
  int sockno;

  // Create normal UDP socket
  if ((sockno = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {perror("Socket Error! "); return(0);}

  // Setup necessary UDP socket stuff
  localaddr.sin_family = AF_INET;
  localaddr.sin_port = htons(port);
  localaddr.sin_addr.s_addr = INADDR_ANY;
  size_localaddr = sizeof(localaddr);


  // Bind socket and get host/port info
  if ((bind(sockno, (struct sockaddr *) &localaddr, sizeof(localaddr))) < 0) {perror("Bind Error! "); return(0);}

  return(sockno);

}

int helper_start_srp_server (int udpsockno) {
  
  struct sockaddr_in peeraddr;
  int size_peeraddr;
  char buffer[256];
  struct srp_socket *srp_sock;

  size_peeraddr = sizeof(peeraddr);

  // Get a ready (nack right now) message from the client to get the address to send to.
  if (recvfrom(udpsockno, buffer, 256, 0, (struct sockaddr *) &peeraddr, &size_peeraddr) < 0) {perror("Receive Error ");return(0);}

  // Create SRP socket and start retransmission listening thread
  srp_sock = srp_server_start(udpsockno, 0, (struct sockaddr *) &peeraddr, sizeof(struct sockaddr));

  return((int) srp_sock);

}

int helper_sendto_srp(int srpsock, int data, int datasize) {
  srp_sendto((struct srp_socket *) srpsock,(void *) data, datasize);
}


int helper_stop_server_srp(int srpsock) {

  srp_server_stop((struct srp_socket *) srpsock);

}
