#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <inttypes.h>
#include "srpclientlib.c"

void lookup_addr(char *, int, char *, struct sockaddr_in *);
int make_sock(struct sockaddr_in, int);
void timestamp(char *, int);


int main(int argc, char **argv) {

  int test, udp_sock;
  struct srp_client_socket *srp_sock; 
  char buffer[SRP_MAX_DATA_SIZE], strcmd[50];
  struct sockaddr_in addr;
  
  struct srp_nack_packet srp_nack;
  int size_localaddr;
  float loss;
  long lat;

  long total_recv, total_good;
 

#ifdef SRP_STAT
  printf("Statistics Version\n");
  srp_stat_init(&stat);
  srp_stat_client_read("client_params.txt");
  
#endif

  if (argc != 4) {
    printf("Usage:  client <host> <port> <tcp/udp/ell/oq>\n");
    exit(1);
  } 
     
  /*
   * All of this socket stuff will be done someday by the user program not by SRP
   */

  /* Lets lookup the address here  */
  printf("Host: %s Port: %d \n",argv[1],atoi(argv[2]));
  lookup_addr(argv[1], atoi(argv[2]), "udp", &addr);
      
  /* setup and connect to address in above line */
  if ((udp_sock = make_sock(addr, SOCK_DGRAM)) == -1) {
    exit(1);
  }

  size_localaddr = sizeof(struct sockaddr_in);


  // Send a message to get the sample test server going
  srp_nack.status = SRP_STATUS_MISSING;
  srp_nack.srp_data_id = 1;
  printf("Sending init message: %d\n",1);
  sendto(udp_sock, &srp_nack, sizeof(struct srp_nack_packet), 0, &addr, sizeof(struct sockaddr_in));
 
  /* set loss and latency values here */
 loss = 0.10;
 lat = 250;


 total_recv = 0;
 total_good = 0;

 srp_sock = srp_client_start(udp_sock, 0, (struct sockaddr *) &addr, sizeof(struct sockaddr_in), loss, lat);
 strcpy(srp_sock->algorithm, argv[3]);


  do {
    test = srp_recvfrom(srp_sock, &buffer, SRP_MAX_DATA_SIZE);
    total_recv++;
    if (test > 0) {
      total_good++;
      timestamp("srpdat.txt", 0);
    } else {
      timestamp("srpdat.txt", 1);
    }

    printf("Received %d (%d) \n",srp_sock->expected, srp_sock->data_pkt_recv);
    printf("Current loss %g \n",srp_sock->ave_loss);
    printf("Current latency %ld \n",srp_sock->ave_lat);
    printf("Current rtt %ld (%ld) \n",srp_sock->rtt, srp_sock->rtt_dev);
    printf("Expect interval %ld (%ld)\n", srp_sock->exp_interval, srp_sock->exp_dev);
    printf("Loss: %g\n", 1.0 - total_good/(float)total_recv);


    /* printf("returned to main\n");*/

  } while(1);
  printf("returned to main\n");
  
  /* We are finished reading from the socket, so close the connection */

  
  srp_client_stop(srp_sock);
  close(udp_sock);

}



void lookup_addr(char *name, int port, char *proto, struct sockaddr_in *sin) {

  struct hostent *addr_info;
  struct servent *port_info;
  char *pAddr, **ppAddr;
  struct in_addr addr;


  if ((addr_info = gethostbyname(name)) == NULL) {perror("Could not get addr! "); exit(1); }
  //  if ((port_info = getservbyport(htons(port), "udp")) == NULL) {perror("Could not get port! "); exit(1); }

  ppAddr = addr_info->h_addr_list;
  pAddr = ppAddr[0];

  memcpy(&addr, pAddr, sizeof(struct in_addr));

  sin->sin_addr = addr;
  sin->sin_family = AF_INET;
  sin->sin_port = htons(port);
  //  sin->sin_port = port_info->s_port;

}


int make_sock(struct sockaddr_in sin, int proto) {

  int socketnum;

  if ((socketnum = socket(AF_INET, proto, 0)) < 0) {perror("Socket Error! "); return(-1);}
  if (connect(socketnum, (struct sockaddr *) &sin, sizeof(sin))) {perror("Connect error! "); return(-1);}


  return(socketnum);

}


void timestamp(char *time_file, int loss) {
  FILE *outfile;

  if ((outfile = fopen(time_file, "at")) == NULL) {
    printf("Error, could not open statistics file \"%s\" for writing\n", time_file);
    exit(1);
  }

  fprintf(outfile, "%ld\t%d\n", gettime(), loss);
  fclose(outfile);

}




















