// SRP Client Library


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
// #include <inttypes.h>
#include <pthread.h>
/* #include "srp_stats.h" */
#include "srp.h"




struct srp_client_socket *srp_client_start(int sock, unsigned int flags, const struct sockaddr *to, int tolen, float loss, int lat) {
  
  struct srp_client_socket *new;

  new = malloc(sizeof(struct srp_client_socket));

  if (new == NULL) {
    return(NULL);
  }

  new->socketnum = sock;
  new->flags = flags;
  memcpy(&(new->to), to, sizeof(struct sockaddr));
  new->tolen = sizeof(struct sockaddr);

  new->expected = 1;
  new->req_loss = loss;
  new->req_lat = lat;
  new->ave_loss = 0.0;
  new->ave_lat = -1;
  new->last_rtt = gettime();
  new->last_probe = gettime();
  new->rtt = -1;
  new->rtt_dev = -1;
  new->lnrt = -1;
  new->last_lnrt = 0;
  new->exp_interval = -1;
  new->exp_dev = -1;
  new->retrans_time = -1;

  return(new);

}



int srp_recvfrom(struct srp_client_socket *socketStr, void *buffer, int max_data_size) {

  struct srp_data_packet srp_in_pkt;
  struct srp_probe_packet probe;
  struct srp_buffer_node *srp_from_list = NULL;
  int size_localaddr;
  char message[20];
  struct timeval tv;
  int retval;
  int num_nacks;
  int temp;
  int still_retrans, need_retrans;


  fd_set readset, writeset, eset;
  struct sockaddr_in peeraddr;

  num_nacks = 0;  

  /* check buffer. 
   * If the expected number is in the list,
   * get the data out, and delete the node
   */
  
  srp_from_list = check_list(&(socketStr->buffer), socketStr->expected);

  if (srp_from_list != NULL) {

    // -------------------
    // Message in buffer
    // -------------------

#ifdef SRP_STAT
    stat.cache_hits++;
#endif

    printf("Cached: %d, %ld, %ld", srp_from_list->msg.data_id, srp_from_list->msg.arrival, srp_from_list->msg.rtt);

    // Update timing
    socketStr->lnrt = srp_from_list->msg.arrival;
    socketStr->last_lnrt = 0;
    update_loss(socketStr, 0);
    update_cache_lat(socketStr, srp_from_list->msg.arrival, srp_from_list->msg.rtt);

    
    

    if (srp_from_list->msg.size > max_data_size) {

      /* the amount read in is greater than the max size */
      memcpy(buffer, srp_from_list->msg.data, max_data_size);
      socketStr->expected = (socketStr->expected+1) % SRP_MAX_DATA_ID;
      free(srp_from_list->msg.data);
      free(srp_from_list);
      
      //printf("from list: max: %d msg:%d\n", max_data_size, srp_from_list->msg.size);
      return(max_data_size);
      
    } else {
      
      /* the packet data is either less than the max, or equal
       * both are handled the same way
       */
      memcpy(buffer, srp_from_list->msg.data, srp_from_list->msg.size);
      socketStr->expected = (socketStr->expected+1) % SRP_MAX_DATA_ID;
      temp = srp_from_list->msg.size;
      free(srp_from_list->msg.data);
      free(srp_from_list); 

      //printf("from list: msg: %d max:%d\n", srp_from_list->msg.size, max_data_size); 
      return(temp);
      
    }
    
  }
  
  
  do {
    
    need_retrans = 0;

    /*
     * Take Care of Probes
     */

    if ((diftime(socketStr->last_rtt, gettime()) > SRP_MAX_UNK_RTT) && (diftime(socketStr->last_probe, gettime()) > SRP_MAX_PROBE_TIME)) {
      probe.srp_head.srp_type = SRP_TYPE_PROBE;
      probe.srp_head.timestamp = gettime();
      socketStr->last_probe = gettime();
      sendto(socketStr->socketnum, &probe, sizeof(struct srp_probe_packet), socketStr->flags, &(socketStr->to), socketStr->tolen); 
    }


    /*
      * Need to do two things here, receive on the socket and store the data in
      * some sort of linked list
      * Also if the needed one has not come in, send a nack
      */
    
    size_localaddr = sizeof(struct sockaddr_in); 
    
    FD_ZERO(&readset);
    FD_ZERO(&writeset);
    FD_ZERO(&eset);
    FD_SET(socketStr->socketnum, &readset);
    FD_SET(socketStr->socketnum, &eset);
    
    tv.tv_sec = 0;
    tv.tv_usec = rto_time(socketStr, num_nacks) * 1000;
    printf("RTO: %ld\n", tv.tv_usec/1000);

    if ((retval = select(socketStr->socketnum + 1, &readset, &writeset, &eset, &tv)) < 0) {
      printf("select failed\n");
      return(EPIPE);
      
    }
    
    if (FD_ISSET(socketStr->socketnum,&eset)) {
      printf("Socket Exception\n");
      return(EPIPE);
    }
    
    if(retval) {
      
      /* There is activity on the socket so read from it */
      if(recvfrom(socketStr->socketnum,&srp_in_pkt, sizeof(struct srp_data_packet), 0, NULL, NULL) < 0) {
	printf("recvfrom Exception\n");
	return(EINVAL);
      }
      
      // If message has time information, read it
      if(srp_in_pkt.srp_head.timestamp != SRP_NO_TIME) {
	update_rtt(socketStr, srp_in_pkt.srp_head.timestamp, gettime());
      }

      // If message is a probe
      if (srp_in_pkt.srp_head.srp_type == SRP_TYPE_PROBE) {
	printf("Received probe: timestamp: %ld\n", srp_in_pkt.srp_head.timestamp);
      }

      // If message is a data message
      if (srp_in_pkt.srp_head.srp_type == SRP_TYPE_DATA) {


#ifdef SRP_STAT
	stat.data_pkts_recv++;
	stat.data_bytes_recv += sizeof(struct srp_packet_header) +  sizeof(struct srp_data_packet_header) + srp_in_pkt.srp_data_head.srp_data_size;
#endif
    

	if (lessthan(srp_in_pkt.srp_data_head.srp_data_id, socketStr->expected)) {
	  
	  // ------------------------------------
	  // Data id less than expected (Too old)
	  // ------------------------------------

	  /* The data_id in the incoming message is less than the expected one
	   * so do nothing with the incoming message.
	   */

#ifdef SRP_STAT
	  stat.too_old++;
#endif
     
	  printf("Got %d want: %d (Too Old)\n", srp_in_pkt.srp_data_head.srp_data_id, socketStr->expected);

	  // Last normal receive time still ok since this is either a packet that was given up on or out of order

	} 
  
	if (greaterthan(srp_in_pkt.srp_data_head.srp_data_id, socketStr->expected)) {
	  
	  // ----------------------------------------
	  // Data id greater than exepected (Too new)
	  // ----------------------------------------

	  /* The data_id in the incoming message is greater than the expected one
	   * so send it to the linked list to be inserted and stored for later on
	   * when the expected one arrives, set up the nack timer to nack for
	   * the missing one again 
	   */

#ifdef SRP_STAT
	  stat.too_new++;
#endif

	  printf("Got %d Want %d(Too New)\n", srp_in_pkt.srp_data_head.srp_data_id, socketStr->expected);
	  // Does not invalidate last normal receive time since out of order is possible

          socketStr->data_pkt_recv++;

	  /* We got one out of order, send a nack right away for the expected one but only if we are just waiting for a normal reception*/
	  if (num_nacks == 0) {
	    need_retrans = 1;
	  }
	  // Ensure that buffer will not fill up too much by not buffering data to far ahead
	  if (lessthanequal(srp_in_pkt.srp_data_head.srp_data_id, (socketStr->expected + SRP_BUFFER_SIZE) % SRP_MAX_DATA_ID)) {
	    add_data(&(socketStr->buffer), srp_in_pkt.srp_data_head.srp_data_id, srp_in_pkt.data, srp_in_pkt.srp_data_head.srp_data_size, gettime(), socketStr->rtt); 
	  }

	}

	if (srp_in_pkt.srp_data_head.srp_data_id == socketStr->expected) {

	  // ---------------------------------------------
	  // Received the expected data id (Good message)
	  // ---------------------------------------------

	  /* If we got here it must be what we expected so return the data to the client */
	  /* we have the packet we want, so lets check the size of the item coming in
	   * against the max_data_size passed into the function
	   * if the data coming in is less than the max_size memcpy what we have and
	   * return the size placed in.
	   * if the data is greater than the max_size only return max_size of data
	   */
	  
	  socketStr->data_pkt_recv++;


	  // If not a retransmission reply
	  if (srp_in_pkt.srp_head.timestamp == SRP_NO_TIME) {

       
	    // Got the expected data id, compute new latency if previous lnrt is valid.
	    if ((socketStr->lnrt >= 0) && (socketStr->last_lnrt == 0)) {
	      update_exp(socketStr, gettime());
	    }
	    socketStr->lnrt = gettime();
	    socketStr->last_lnrt = 0;
	    printf("Got %d (Normal)\n", srp_in_pkt.srp_data_head.srp_data_id);
	  } else {
	    socketStr->last_lnrt++;
	    printf("Got %d (Good (Retrans))\n", srp_in_pkt.srp_data_head.srp_data_id);
	  }

	  update_loss(socketStr, 0);
	  update_lat(socketStr);


	  if (srp_in_pkt.srp_data_head.srp_data_size > max_data_size) {
	    /* the amount read in is greater than the max size */
	    memcpy((struct srp_data_packet *)buffer, srp_in_pkt.data, max_data_size);
	    socketStr->expected = (socketStr->expected+1) % SRP_MAX_DATA_ID;
	    // printf("From sock: max: %d, msg: %d\n", max_data_size, srp_in_pkt.srp_data_head.srp_data_size);
	    return(max_data_size);
	    
	  } else {
	  
	    /* the packet data is either less than the max, or equal
	     * both are handled the same way
	     */
	    memcpy(buffer, srp_in_pkt.data, srp_in_pkt.srp_data_head.srp_data_size);
	    socketStr->expected = (socketStr->expected+1) % SRP_MAX_DATA_ID;
	    // printf("From sock: msg: %d, max: %d\n", srp_in_pkt.srp_data_head.srp_data_size, max_data_size);
	    return(srp_in_pkt.srp_data_head.srp_data_size);
	  
	  }
	}
      }
    } else {

      // Timeout

      /* Timer has expired call for nack for the expected */
      need_retrans = 1;
      printf("Timeout want: %d\n", socketStr->expected);

    }
    if (need_retrans) {
      still_retrans = should_retransmit(socketStr, num_nacks);
    } else {
      still_retrans = 1;
    }

    if (need_retrans && still_retrans) {
      printf("##Retrans##\n");
      socketStr->retrans_time = gettime();
      send_nack(socketStr->expected, socketStr);
      num_nacks++;
    }


  } while(still_retrans);

  // -----------------------------
  // Done Retransmitting (Give up)
  // ----------------------------

#ifdef SRP_STAT
  stat.give_ups++;
#endif

  // lnrt is now one data id older
  socketStr->last_lnrt++;

  socketStr->give_up++;
  update_loss(socketStr, 1);
  socketStr->expected = (socketStr->expected+1) % SRP_MAX_DATA_ID;
  printf("*Gave up*\n");
  return(0);
  
}



void send_nack(int missing, struct srp_client_socket *socket) {
 
  struct srp_nack_packet srp_nack;

  srp_nack.srp_head.srp_type = SRP_TYPE_NACK;
  srp_nack.srp_head.timestamp = gettime();
  srp_nack.status = SRP_STATUS_MISSING;
  srp_nack.srp_data_id = missing;
  printf("Sending Nack for: %d\n", missing);


#ifdef SRP_STAT
  stat.retrans_rqstd++;
  stat.retrans_bytes_rqstd += sizeof(struct srp_nack_packet);
#endif

      sendto(socket->socketnum, &srp_nack, sizeof(struct srp_nack_packet), socket->flags, &(socket->to), socket->tolen); 
}


int add_data(struct srp_buffer_node **mm_data, int data_id, char *data, int data_size, long arrival, long rtt) { 

  int i;
  struct srp_buffer_node  *new,*temp,*cur;

  if (*mm_data == NULL) { 

    // Only item in list
    new = malloc(sizeof(struct srp_buffer_node));
    if (new == NULL) {
      printf("ERROR cannot malloc for linked list\n");
      return(ENOMEM);
    }

    new->msg.data_id = data_id;
    new->msg.size = data_size;
    new->msg.arrival = arrival;
    new->msg.rtt = rtt;

    if (data_size > SRP_MAX_DATA_SIZE) {
      // Error, Some data will be cut off
      new->msg.data = malloc(SRP_MAX_DATA_SIZE);
      if (new->msg.data == NULL) {
	printf("Out of buffer space\n");
	return(ENOMEM);
      }
      memcpy(new->msg.data, data, SRP_MAX_DATA_SIZE);
    } else {
      new->msg.data = malloc(data_size);
      if (new->msg.data == NULL) {
	printf("Out of buffer space\n");
	return(-1);
      }
      memcpy(new->msg.data, data, data_size);   
    }
    new->next = NULL;
    *mm_data = new;

  } else {

     if((*mm_data)->msg.data_id == data_id) return;
	
     if (greaterthan((*mm_data)->msg.data_id, data_id)) {

       new = malloc(sizeof(struct srp_buffer_node));
       if (new == NULL) {
	 printf("ERROR cannot malloc for linked list\n");
	 return(ENOMEM);
       }
       new->msg.data_id = data_id;
       new->msg.size = data_size;
       new->msg.arrival = arrival;
       new->msg.rtt = rtt;
       
       if (data_size > SRP_MAX_DATA_SIZE) {
	 // Error, Some data will be cut off
	 new->msg.data = malloc(SRP_MAX_DATA_SIZE);
	 if (new->msg.data == NULL) {
	   printf("Out of buffer space\n");
	   return(ENOMEM);
	 }
	 memcpy(new->msg.data, data, SRP_MAX_DATA_SIZE);
       } else {
	 new->msg.data = malloc(data_size);
	 if (new->msg.data == NULL) {
	   printf("Out of buffer space\n");
	   return(ENOMEM);
	 }
	 memcpy(new->msg.data, data, data_size);   
       }     
 
       new->next = (*mm_data);
       (*mm_data) = new;
       return;
     } else {

       /* somewhere in the middle */

       for (cur = *mm_data; cur->next != NULL; cur=cur->next) {
       
	 /* check and see if goes before */
	 if(cur->next->msg.data_id == data_id) return;
 
	 if (greaterthan(cur->next->msg.data_id, data_id)) break;
        
       } /* for loop */

       new = malloc(sizeof(struct srp_buffer_node));
      if (new == NULL) {
	 printf("ERROR cannot malloc for linked list\n");
	 return(ENOMEM);
       }
       new->msg.data_id = data_id;
       new->msg.size = data_size;
       new->msg.arrival = arrival;
       new->msg.rtt = rtt;


       if (data_size > SRP_MAX_DATA_SIZE) {
	 // Error, Some data will be cut off
	 new->msg.data = malloc(SRP_MAX_DATA_SIZE);
	 if (new->msg.data == NULL) {
	   printf("Out of buffer space\n");
	   return(ENOMEM);
	 }
	 memcpy(new->msg.data, data, SRP_MAX_DATA_SIZE);
       } else {
	 new->msg.data = malloc(data_size);
	 if (new->msg.data == NULL) {
	   printf("Out of buffer space\n");
	   return(ENOMEM);
	 }
	 memcpy(new->msg.data, data, data_size);   
       }      
       new->next = cur->next;
       cur->next = new;
     }
  }
} /* closes function */



void print_data(struct srp_buffer_node *newest) {
  struct srp_buffer_node *cur;
  char *tmp;
  int i;

  for (cur = newest; cur != NULL; cur=cur->next) {
    printf("Frame %d: ",cur->msg.data_id);
    tmp = (char *) cur->msg.data;
      for (i=0;i<cur->msg.size;i++) {
	printf("%c", tmp[i]);
      }
    printf("\n--------------\n");
  }

}


struct srp_buffer_node *check_list(struct srp_buffer_node **newest,int expid) {

  struct srp_buffer_node *cur, *temp;
 
  if (*newest == NULL) {
    return(NULL);
  } else {
    if ((*newest)->msg.data_id == expid) {
      cur = (*newest);
      (*newest) = cur->next;
      return(cur);
    }
  }
  return(NULL); 

}

void srp_client_stop(struct srp_client_socket *sock) {

  struct srp_buffer_node *cur, *freeing;

  // Free data buffer
  cur=sock->buffer;
  while (cur != NULL) {
    freeing = cur;
    cur=cur->next;
    free(freeing->msg.data);
    free(freeing);
  }

   free(sock);

}


int update_exp(struct srp_client_socket *sock, long cur_time) {
 
  long exp_int;

  if (sock->lnrt >= 0) {
    exp_int = diftime(sock->lnrt, cur_time);
    
    if (sock->exp_interval < 0) {
      sock->exp_interval = exp_int;
    } else {
      if (sock->exp_dev < 0) {
	sock->exp_dev = abs(exp_int - sock->exp_interval);
      } else {
	sock->exp_dev = SRP_EXPDEV_SMOOTH * sock->exp_dev + (1.0 - SRP_EXPDEV_SMOOTH) * abs(exp_int - sock->exp_interval);
      }
      sock->exp_interval = SRP_EXP_SMOOTH * sock->exp_interval + (1.0 - SRP_EXP_SMOOTH) * exp_int;
    }
  }

  return(0);

}

int update_loss(struct srp_client_socket *sock, int gave_up) {
 
  if (sock->ave_loss < 0) {
    sock->ave_loss = gave_up;
  } else {
    sock->ave_loss = SRP_LOSS_SMOOTH * sock->ave_loss + (1.0 - SRP_LOSS_SMOOTH) * gave_up;
  }

  return(0);

}


int update_rtt(struct srp_client_socket *sock, long probe_time, long cur_time) {
 
  long int cur_rtt;

  sock->last_rtt = gettime();

  cur_rtt = diftime(probe_time, cur_time);

  if (sock->rtt < 0) {
    sock->rtt = cur_rtt;
  } else {
    if (sock->rtt_dev <= 0) {
      sock->rtt_dev = abs(cur_rtt - sock->rtt);
    } else {
      sock->rtt_dev = SRP_RTTDEV_SMOOTH * sock->rtt_dev + (1.0 - SRP_RTTDEV_SMOOTH) * abs(cur_rtt - sock->rtt);
    }
    sock->rtt = SRP_RTT_SMOOTH * sock->rtt + (1.0 - SRP_RTT_SMOOTH) * cur_rtt;
  }


  return(0);

}


int update_lat(struct srp_client_socket *sock) {
 
  long int cur_lat, exp_time;
  
  if ((sock->lnrt >= 0) && (sock->exp_interval >= 0)) {

    exp_time = sock->lnrt + sock->last_lnrt * sock->exp_interval;

    if (lessthanequal_time(exp_time, gettime())) { 

      cur_lat = diftime(exp_time, gettime()) + sock->rtt/2;
      
      if (sock->ave_lat < 0) {
	sock->ave_lat = cur_lat;
      } else {
	sock->ave_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * cur_lat;
      }
    

    } else {
      printf("ERROR in latency, current time is sooner than time expected (%ld ,%ld )\n", exp_time, gettime());
    }

  }

  return(0);

}

int update_cache_lat(struct srp_client_socket *sock, long arrival, long rtt) {
 
  long int cur_lat;

  if (rtt >= 0) {

    cur_lat = diftime(arrival, gettime()) + rtt/2;
 
    if (sock->ave_lat < 0) {
      sock->ave_lat = cur_lat;
    } else {
      sock->ave_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * cur_lat;
    }

  }

  return(0);

}

int should_retransmit(struct srp_client_socket *sock, long num_nacks) {
  // Return 1 to Retransmit if needed or 0 to give up

  // TCP
  if (strcmp(sock->algorithm, "tcp") == 0) { 
    return(1);
  }

  // UDP
  if (strcmp(sock->algorithm, "udp") == 0) { 
    return(0);
  }

  // After SRP_NUM_RETRIES retransmittions
  // return(num_nacks < SRP_NUM_RETRIES);

  // Based on projected loss and latency evening out loss and latency
  if (strcmp(sock->algorithm, "ell") == 0) {
    
    long current_lat, exp_time;
    float proj_loss;
    long proj_lat;
    float loss_pct;
    float lat_pct;
    float diff_lat;
    float diff_loss;

    if ((sock->ave_lat < 0) || (sock->lnrt < 0) || (sock->rtt < 0) || (sock->exp_interval < 0)) {return(1);}

    exp_time = sock->lnrt + sock->last_lnrt * sock->exp_interval;
  
    if (lessthanequal_time(exp_time, gettime())) {
      current_lat = diftime(exp_time, gettime()) + sock->rtt/2;


      // Projected for giving up
      proj_loss = SRP_LOSS_SMOOTH * sock->ave_loss + (1.0 - SRP_LOSS_SMOOTH);
      proj_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * current_lat;
      
      loss_pct = proj_loss/sock->req_loss;
      lat_pct = proj_lat/(float) sock->req_lat;
      diff_loss = abs(loss_pct - lat_pct);
    
      printf("Decide: Loss: %g (%g%%) %ld (%g%%), ", proj_loss, loss_pct * 100, proj_lat, lat_pct * 100);

      // Projected for retrans
      proj_loss = SRP_LOSS_SMOOTH * sock->ave_loss;
      proj_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * (current_lat + sock->rtt);

      loss_pct = proj_loss/sock->req_loss;
      lat_pct = proj_lat/(float) sock->req_lat;
      diff_lat = abs(loss_pct - lat_pct);
    
      printf("Lat: %g (%g%%) %ld (%g%%), %d\n", proj_loss, loss_pct * 100, proj_lat, lat_pct * 100, diff_loss > diff_lat);
  

      return(diff_loss > diff_lat);
    } else {
      printf("ERROR in decision, current time is sooner than time expected (%ld, %ld, %d)\n", exp_time, gettime(), sock->last_lnrt);
      return(1);
    }

  }

  // Based on projected loss and latency closest to optimum
  if (strcmp(sock->algorithm, "oq") == 0) { 
 
    long current_lat, exp_time;
    float proj_loss;
    long proj_lat;
    float loss_pct;
    float lat_pct;
    float diff_lat;
    float diff_loss;

    if ((sock->ave_lat < 0) || (sock->lnrt < 0) || (sock->rtt < 0) || (sock->exp_interval < 0)) {return(1);}
    
    exp_time = sock->lnrt + sock->last_lnrt * sock->exp_interval;
    
    if (lessthanequal_time(exp_time, gettime())) {
      current_lat = diftime(exp_time, gettime()) + sock->rtt/2;
      
      
      // Projected for giving up
      proj_loss = SRP_LOSS_SMOOTH * sock->ave_loss + (1.0 - SRP_LOSS_SMOOTH);
      proj_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * current_lat;
      
      loss_pct = proj_loss/sock->req_loss;
      lat_pct = proj_lat/(float) sock->req_lat;
      diff_loss = loss_pct * loss_pct + lat_pct * lat_pct;
      
      printf("Decide: Loss: %g (%g%%) %ld (%g%%), ", proj_loss, loss_pct * 100, proj_lat, lat_pct * 100);
      
      // Projected for retrans
      proj_loss = SRP_LOSS_SMOOTH * sock->ave_loss;
      proj_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * (current_lat + sock->rtt);
      
      loss_pct = proj_loss/sock->req_loss;
      lat_pct = proj_lat/(float) sock->req_lat;
      diff_lat = loss_pct * loss_pct + lat_pct * lat_pct;
      
      printf("Lat: %g (%g%%) %ld (%g%%), %d\n", proj_loss, loss_pct * 100, proj_lat, lat_pct * 100, diff_loss > diff_lat);
      
      
      return(diff_loss > diff_lat);
    } else {
      printf("ERROR in decision, current time is sooner than time expected (%ld, %ld, %d)\n", exp_time, gettime(), sock->last_lnrt);
      return(1);
    }
  }
 
  // Based on projected loss and latency evening out loss and latency
  if (strcmp(sock->algorithm, "ell2") == 0) {
    
    long current_lat, exp_time;
    float proj_loss;
    long proj_lat;
    float loss_pct;
    float lat_pct;
    float diff_lat;
    float diff_loss;


    // Worst case 
    //if ((sock->rtt + 4 * sock->rtt_dev) > sock->req_lat) {
    //  printf("Retrans cost too high\n");
    //  return(0);
    //}

    if ((sock->ave_lat < 0) || (sock->lnrt < 0) || (sock->rtt_dev < 0) || (sock->rtt < 0) || (sock->exp_interval < 0)) {return(1);}

    exp_time = sock->lnrt + sock->last_lnrt * sock->exp_interval;
  
    if (lessthanequal_time(exp_time, gettime())) {
      current_lat = diftime(exp_time, gettime()) + sock->rtt/2;


      // Projected for giving up
      proj_loss = SRP_LOSS_SMOOTH * sock->ave_loss + (1.0 - SRP_LOSS_SMOOTH);
      proj_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * current_lat;
      
      loss_pct = proj_loss/sock->req_loss;
      lat_pct = proj_lat/(float) sock->req_lat;
      diff_loss = abs(loss_pct - lat_pct);
    
      printf("Decide: Loss: %g (%g%%) %ld (%g%%), ", proj_loss, loss_pct * 100, proj_lat, lat_pct * 100);

      // Projected for retrans
      proj_loss = SRP_LOSS_SMOOTH * sock->ave_loss;
      proj_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * (current_lat + sock->rtt + 4 * sock->rtt_dev);

      loss_pct = proj_loss/sock->req_loss;
      lat_pct = proj_lat/(float) sock->req_lat;
      diff_lat = abs(loss_pct - lat_pct);
    
      printf("Lat: %g (%g%%) %ld (%g%%), %d\n", proj_loss, loss_pct * 100, proj_lat, lat_pct * 100, diff_loss > diff_lat);
  

      return(diff_loss > diff_lat);
    } else {
      printf("ERROR in decision, current time is sooner than time expected (%ld, %ld, %d)\n", exp_time, gettime(), sock->last_lnrt);
      return(1);
    }

  }

  // Based on projected loss and latency closest to optimum
  if (strcmp(sock->algorithm, "oq2") == 0) { 
 
    long current_lat, exp_time;
    float proj_loss;
    long proj_lat;
    float loss_pct;
    float lat_pct;
    float diff_lat;
    float diff_loss;

    if ((sock->ave_lat < 0) || (sock->lnrt < 0) || (sock->rtt_dev < 0) || (sock->rtt < 0) || (sock->exp_interval < 0)) {return(1);}
    
    exp_time = sock->lnrt + sock->last_lnrt * sock->exp_interval;
    
    if (lessthanequal_time(exp_time, gettime())) {
      current_lat = diftime(exp_time, gettime()) + sock->rtt/2;
      
      
      // Projected for giving up
      proj_loss = SRP_LOSS_SMOOTH * sock->ave_loss + (1.0 - SRP_LOSS_SMOOTH);
      proj_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * current_lat;
      
      loss_pct = proj_loss/sock->req_loss;
      lat_pct = proj_lat/(float) sock->req_lat;
      diff_loss = loss_pct * loss_pct + lat_pct * lat_pct;
      
      printf("Decide: Loss: %g (%g%%) %ld (%g%%), ", proj_loss, loss_pct * 100, proj_lat, lat_pct * 100);
      
      // Projected for retrans
      proj_loss = SRP_LOSS_SMOOTH * sock->ave_loss;
      proj_lat = SRP_LAT_SMOOTH * sock->ave_lat + (1.0 - SRP_LAT_SMOOTH) * (current_lat + sock->rtt + 4 * sock->rtt_dev);
      
      loss_pct = proj_loss/sock->req_loss;
      lat_pct = proj_lat/(float) sock->req_lat;
      diff_lat = loss_pct * loss_pct + lat_pct * lat_pct;
      
      printf("Lat: %g (%g%%) %ld (%g%%), %d\n", proj_loss, loss_pct * 100, proj_lat, lat_pct * 100, diff_loss > diff_lat);
      
      
      return(diff_loss > diff_lat);
    } else {
      printf("ERROR in decision, current time is sooner than time expected (%ld, %ld, %d)\n", exp_time, gettime(), sock->last_lnrt);
      return(1);
    }
  }


}

long rto_time(struct srp_client_socket *sock, int num_nacks) {
  // Return number of milliseconds until timeout

  /*
   printf("lnrt: %ld (%d)", sock->lnrt, sock->last_lnrt); 
   return(SRP_RTO);
  */

  
  long expected_time;
    
  printf("lnrt: %ld (%d)", sock->lnrt, sock->last_lnrt);

  if (num_nacks > 0) {
    if (sock->retrans_time < 0) {
      printf("Error in RTO: retrans_time not set!\n");
      return(SRP_RTO);
    }
    if ((sock->rtt < 0) || (sock->rtt_dev < 0)) {
      return(SRP_RTO);
    } else {
      
      expected_time = sock->retrans_time + sock->rtt + 4 * sock->rtt_dev;

      if (lessthanequal_time(expected_time, gettime())) {
	return(0); // Past time that retrans should appear, timeout immediatly
      } else {
	return(diftime(gettime(), expected_time));
      }

    }

  } else {
  
    if ((sock->exp_interval < 0) || (sock->exp_dev < 0) || (sock->lnrt < 0)) {
      return(SRP_RTO);
    } else {
      expected_time = sock->lnrt + (1 + sock->last_lnrt) * sock->exp_interval;
      if (lessthanequal_time(expected_time, gettime())) {
	return(0); // Retransmit immediately since we missed expected time
      } else {
	return(diftime(gettime(), expected_time) + 4 * sock->exp_dev);
      }
    }
  }
  
 

}


int lessthan(int id1, int id2) {
  
  int lowlim, highlim;

  lowlim = SRP_MAX_DATA_ID / 4;
  highlim = lowlim * 3;

  if ((id1 > highlim) && (id2 < lowlim)) {
    return(1);
  } 
  if ((id2 > highlim) && (id1 < lowlim)) {
    return(0);
  } 

  return(id1<id2);
  
}

int lessthanequal(int id1, int id2) {
  
  int lowlim, highlim;

  if (id1 == id2) return(1);

  lowlim = SRP_MAX_DATA_ID / 4;
  highlim = lowlim * 3;

  if ((id1 > highlim) && (id2 < lowlim)) {
    return(1);
  } 
  if ((id2 > highlim) && (id1 < lowlim)) {
    return(0);
  } 

  return(id1<id2);
  
}

int greaterthan(int id2, int id1) {
  
  int lowlim, highlim;

  lowlim = SRP_MAX_DATA_ID / 4;
  highlim = lowlim * 3;

  if ((id1 > highlim) && (id2 < lowlim)) {
    return(1);
  } 
  if ((id2 > highlim) && (id1 < lowlim)) {
    return(0);
  } 

  return(id1<id2);
  
}

int greaterthanequal(int id2, int id1) {
  
  int lowlim, highlim;

  if (id1 == id2) return(1);

  lowlim = SRP_MAX_DATA_ID / 4;
  highlim = lowlim * 3;

  if ((id1 > highlim) && (id2 < lowlim)) {
    return(1);
  } 
  if ((id2 > highlim) && (id1 < lowlim)) {
    return(0);
  } 

  return(id1<id2);
  
}

long diftime(long old_time, long new_time) {
  long dif_time;
  dif_time = new_time - old_time;
  
  if(dif_time < 0)
    return(1000000 + dif_time);
  else
    return dif_time;
}

long gettime() {
  struct timeval cur_time;
  struct timezone dummy;

  if (gettimeofday(&cur_time, &dummy) >= 0)
    return((cur_time.tv_sec % 1000) * 1000) + (cur_time.tv_usec / 1000);
  else
    return(-1);
}


int lessthanequal_time(long id1, long id2) {
  
  int lowlim, highlim;

  if (id1 == id2) return(1);

  lowlim = 1000000 / 4;
  highlim = lowlim * 3;

  if ((id1 > highlim) && (id2 < lowlim)) {
    return(1);
  } 
  if ((id2 > highlim) && (id1 < lowlim)) {
    return(0);
  } 

  return(id1<id2);
  
}

void reset(struct srp_client_socket *sock, int expected, long lnrt) {
  

  printf("!!!!!!!! RESET !!!!!!!!!\n");
  sock->expected = expected;
  sock->lnrt = lnrt;
  sock->last_lnrt = 0;
  sock->ave_lat = 0;
  sock->ave_loss = 0.0;
  
}

